/*---------------------------------------------------------------------*/
/*   This program is free software; you can redistribute it            */
/*   and/or modify it under the terms of the GNU General Public        */
/*   License as published by the Free Software Foundation; either      */
/*   version 2 of the License, or (at your option) any later version.  */
/*                                                                     */
/*   This program is distributed in the hope that it will be useful,   */
/*   but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the     */
/*   GNU General Public License for more details.                      */
/*                                                                     */
/*   You should have received a copy of the GNU General Public         */
/*   License along with this program; if not, write to the Free        */
/*   Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,   */
/*   MA 02111-1307, USA.                                               */
/*---------------------------------------------------------------------*/
/*=====================================================================*/
/*    serrano/prgm/project/bglk/runtime/Clib/tclvar.c                  */
/*    -------------------------------------------------------------    */
/*    Author      :  Manuel Serrano                                    */
/*    Creation    :  Fri Oct 16 14:04:08 1998                          */
/*    Last change :  Thu Apr 29 16:48:47 1999 (serrano)                */
/*    -------------------------------------------------------------    */
/*    TK variable handling                                             */
/*=====================================================================*/
#include <bigloo.h>
#include <tclInt.h>
#include <tclglue.h>
#include <assert.h>
#define TRACING 	(1<<20)

/*---------------------------------------------------------------------*/
/*    struct traceinfo                                                 */
/*---------------------------------------------------------------------*/
typedef struct traceinfo {
   /* Procedure to call when operations given */
   /* by flags are performed on variable.     */
   Tcl_VarTraceProc *traceProc; 
   /* Argument to pass to proc.               */
   ClientData clientData;
   /* What events the trace procedure is      */
   /* interested in OR-ed combination of:     */
   /*  TCL_TRACE_READS, TCL_TRACE_WRITES, and */
   /*  TCL_TRACE_UNSETS.                      */
   int flags;
   /* Next in list of traces associated with  */
   /* a particular variable.                  */
   struct traceinfo *nextPtr;   
} traceinfo;
 
/*---------------------------------------------------------------------*/
/*    struct varinfo                                                   */
/*---------------------------------------------------------------------*/
typedef struct varinfo {
   /* Value of this variable                  */
   obj_t value;
   /* The trace (if any) associated to that   */
   /* variabe                                 */
   struct traceinfo *info;
} varinfo;

/*---------------------------------------------------------------------*/
/*    static Tcl_HashTable VarTable ...                                */
/*    -------------------------------------------------------------    */
/*    Global hash table retaining variables                            */
/*---------------------------------------------------------------------*/
static Tcl_HashTable VarTable;

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    tcl_lookup_command ...                                           */
/*---------------------------------------------------------------------*/
obj_t
tcl_lookup_command( char *name ) {
   Tcl_HashEntry *entry;
   
   if( entry = Tcl_FindHashEntry(&(((Interp *)SCM_main_interp)->globalNsPtr->cmdTable), name ) ) {
      return (obj_t)Tcl_GetHashValue(entry);
   } else {
      return BUNSPEC;
   }
}

/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    tk_makevar ...                                                   */
/*---------------------------------------------------------------------*/
static int
tk_makevar( char *name, int err ) {
   obj_t value;
   Tcl_HashEntry *entry;
   int new;
   varinfo *vi;

   entry = Tcl_CreateHashEntry( &VarTable, name, &new );

   if( new ) {
      vi = (varinfo *)ckalloc( sizeof( struct varinfo ) );

      vi->info  = 0L;
      vi->value = BUNSPEC;
      Tcl_SetHashValue( entry, vi );
   } else {
      if( err ) 
	 C_FAILURE( "tk-make-variable",
		    "Illegal variable redefinition",
		    string_to_bstring( name ) );
   }
   
   return 1;
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    SCM_tk_makevar ...                                               */
/*    -------------------------------------------------------------    */
/*    Defines a variable and warn if that variable already exists.     */
/*---------------------------------------------------------------------*/
int
SCM_tk_makevar( char *name ) {
   return tk_makevar( name, 1 );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    SCM_tk_makevar_silent ...                                        */
/*    -------------------------------------------------------------    */
/*    Define a variable without warning if the variable already        */
/*    exists.                                                          */
/*---------------------------------------------------------------------*/
int
SCM_tk_makevar_silent( char *name ) {
   return tk_makevar( name, 0 );
}

/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    change_value ...                                                 */
/*    -------------------------------------------------------------    */
/*    This function is called by Scheme when a there's a change on a   */
/*    traced global variable (using a set! or a define).               */
/*---------------------------------------------------------------------*/
static void change_value( traceinfo *ti, char *name ) {
  traceinfo *p;

  for( p = ti; p ; p = p->nextPtr ) {
     /* Invoke trace procedure if not already active */
     if (p->flags & TRACING)
	continue;

     p->flags |= TRACING;

     (*p->traceProc)( p->clientData, SCM_main_interp, name, "", p->flags );
     /* Unset our flag */
     p->flags &= ~TRACING;
  }
}

/*---------------------------------------------------------------------*/
/*    obj_t                                                            */
/*    SCM_tk_scm_getvar ...                                            */
/*---------------------------------------------------------------------*/
obj_t
SCM_tk_scm_getvar( char *name ) {
  Tcl_HashEntry *entry;

 loop:
  if( entry = Tcl_FindHashEntry( &VarTable, name ) ) {
     varinfo *vi = (varinfo *)Tcl_GetHashValue( entry );

     return vi->value;
  } else {
     tk_makevar( name, 0 );
     goto loop;
  }
}  

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    SCM_tcl_getvar ...                                               */
/*---------------------------------------------------------------------*/
char *
SCM_tcl_getvar( char *name, char *env ) {
   assert( name );
   return SCM_scm_to_tk_string( SCM_tk_scm_getvar( name ) );
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    SCM_tcl_getvar2 ...                                              */
/*---------------------------------------------------------------------*/
char *
SCM_tcl_getvar2( char *name1, char *name2, char *env ) {
   if (name2 && *name2) {
      char *res, *s = (char *)alloca( strlen(name1) + strlen(name2) + 3 );

      sprintf(s, "%s{%s}", name1, name2);
      res = SCM_tcl_getvar( s, env );
      return res;
   }
   return SCM_tcl_getvar( name1, env );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    SCM_tk_scm_setvar ...                                            */
/*---------------------------------------------------------------------*/
int
SCM_tk_scm_setvar( char *name, obj_t value ) {
   Tcl_HashEntry *entry;
   varinfo *vi;

 loop:
   entry = Tcl_FindHashEntry( &VarTable, name );

   if( !entry ) {
      tk_makevar( name, 0 );
      goto loop;
   } else {
      vi = (varinfo *)Tcl_GetHashValue( entry );

      vi->value = value;
   
      if( vi->info )
	 change_value( vi->info, name );
   }
   return 1;
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    SCM_tcl_setvar ...                                               */
/*---------------------------------------------------------------------*/
char *
SCM_tcl_setvar( char *name, char *val, int flags, char *env ) {
   obj_t value;
   Tcl_HashEntry *entry;
   varinfo *vi;
 loop:
   if( !(entry = Tcl_FindHashEntry( &VarTable, name )) ) {
      tk_makevar( name, 0 );
      goto loop;
   } else {
      if( flags & SCM_STRINGIFY ) {
	 /* Val is already a string, since it comes from Tk */
	 value = string_to_bstring( val );
      } else {
	 if( *val ) {
	    value = SCM_tk_string_to_scm( val );
	    if( value == BUNSPEC ) return NULL;
	 } else 
	    value = c_constant_string_to_string( "" );
      }

/*       printf( "setvar:name[%s] val:[%s] value:", name, val );       */
/*       dprint( value );                                              */
      
      vi = (varinfo *)Tcl_GetHashValue( entry );
      vi->value = value;

      if( vi->info )
	 change_value( vi->info, name );
      return val;
   }
}

/*---------------------------------------------------------------------*/
/*    char *                                                           */
/*    SCM_tcl_setvar2 ...                                              */
/*---------------------------------------------------------------------*/
char *
SCM_tcl_setvar2( char *name1, char *name2, char *val, int flags, char *env ) {
   if( name2 && *name2 ) {
      char *res, *s = (char *)alloca( strlen(name1) + strlen(name2) + 3 );

      sprintf( s, "%s{%s}", name1, name2 );
      res = SCM_tcl_setvar( s, val, flags, env );
      return res;
   }
   return SCM_tcl_setvar( name1, val, flags, env );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    SCM_tk_variable_exists_p ...                                     */
/*---------------------------------------------------------------------*/
int
SCM_tk_variable_exists_p( char *name ) {
    return (int)Tcl_FindHashEntry( &VarTable, name );
}
   
/*---------------------------------------------------------------------*/
/*    static int                                                       */
/*    trace_var ...                                                    */
/*---------------------------------------------------------------------*/
static int
trace_var( Tcl_Interp *interp,
	   char *var,
	   int flags,
	   Tcl_VarTraceProc *proc,
	   ClientData clientData ) {
   Tcl_HashEntry *entry;
   traceinfo	 *ti;
   varinfo       *vi;

   if( !(entry = Tcl_FindHashEntry( &VarTable, var )) ) {
      fprintf( stderr, "*** WARNING:tracevar:Variable unbound -- %s\n", var );
      interp->result = "Variable unbound";
      return TCL_ERROR;
   } else {
      vi = (varinfo *)Tcl_GetHashValue( entry );
     
      /* Create the value associated to the "var" key */
      ti= (traceinfo *) ckalloc( sizeof (traceinfo) );
      /* Unset has no meaning in stk */;
      ti->flags	 = flags  & ~TCL_TRACE_UNSETS; 
      ti->traceProc  = proc;
      ti->clientData = clientData;
      ti->nextPtr    = vi->info;
  
      vi->info       = ti;
  
      return TCL_OK;
   }
}
 
/*---------------------------------------------------------------------*/
/*    static void                                                      */
/*    untrace_var ...                                                  */
/*---------------------------------------------------------------------*/
static void
untrace_var( Tcl_Interp *interp,
	     char *var,
	     int flags,
	     Tcl_VarTraceProc *proc,
	     ClientData clientData ) {
   Tcl_HashEntry *entry;
   traceinfo *p, *prev;
   varinfo   *vi;

   if( entry = Tcl_FindHashEntry( &VarTable, var) ) {
      /* Variable is traced. Try to find correponding trace function */
      /* Unset has no meaning for us                                 */
      flags &= ~TCL_TRACE_UNSETS;

/*       printf( "untrace_var [%s]\n", var );                          */
      vi = ((varinfo *) Tcl_GetHashValue( entry ));
      p  = vi->info;
      
      for( prev=NULL; p ; prev=p, p=p->nextPtr ) {
	 if( p->traceProc == proc &&
	     p->flags == flags    &&
	     p->clientData == clientData )
	break;
      }
      if( p ) {
	 if( prev == NULL )
	    vi->info = p->nextPtr;
	 else
	    prev->nextPtr = p->nextPtr;
      }
   } else
      fprintf( stderr, "*** WARNING: Variable unbound -- %s\n", var );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    Tcl_TraceVar ...                                                 */
/*---------------------------------------------------------------------*/
int
Tcl_TraceVar( Tcl_Interp *interp,
	      char *var,
	      int flags,
	      Tcl_VarTraceProc *proc,
	      ClientData clientData ) {
  return trace_var( interp, var, flags, proc, clientData );
}

/*---------------------------------------------------------------------*/
/*    int                                                              */
/*    Tcl_TraceVar2 ...                                                */
/*---------------------------------------------------------------------*/
int
Tcl_TraceVar2( Tcl_Interp *interp,
	       char *name1,
	       char *name2,
	       int flags,
	       Tcl_VarTraceProc *proc,
	       ClientData clientData ) {
  return trace_var( interp, name1, flags, proc, clientData );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    Tcl_UntraceVar ...                                               */
/*---------------------------------------------------------------------*/
void
Tcl_UntraceVar( Tcl_Interp *interp,
		char *var,
		int flags,
		Tcl_VarTraceProc *proc,
		ClientData clientData ) {
  untrace_var( interp, var, flags, proc, clientData );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    Tcl_UntraceVar2 ...                                              */
/*---------------------------------------------------------------------*/
void
Tcl_UntraceVar2( Tcl_Interp *interp,
		 char *name1,
		 char *name2,
		 int flags,
		 Tcl_VarTraceProc *proc,
		 ClientData clientData ) {
  untrace_var( interp, name1, flags, proc, clientData );
}

/*---------------------------------------------------------------------*/
/*    void                                                             */
/*    SCM_tcl_init_tracevar ...                                        */
/*---------------------------------------------------------------------*/
void
SCM_tcl_init_tracevar( void ) {
#if( (ckalloc == GC_malloc) && (ckalloc == GC_MALLOC) )
   GC_add_roots( &VarTable, &VarTable + 1 );
#endif
   Tcl_InitHashTable( &VarTable, TCL_STRING_KEYS );
}
